from dexy.node import Node
from dexy.node import DocNode
from dexy.node import PatternNode
from dexy.task import Task
from dexy.tests.utils import wrap
from dexy.wrapper import Wrapper
import time

def test_node_caching():
    with wrap() as wrapper:
        with open("hello.py", "w") as f:
            f.write("print 1+2\n")

        with open("doc.txt", "w") as f:
            f.write("1 + 1 = {{ d['hello.py|py'] }}")

        hello_py = DocNode("hello.py|py", wrapper=wrapper)

        doc_txt = DocNode("doc.txt|jinja",
                inputs = [hello_py],
                wrapper=wrapper)

        wrapper.run_docs(doc_txt)

        hello_py_doc = hello_py.children[0]
        doc_txt_doc = doc_txt.children[0]

        assert str(doc_txt_doc.output()) == "1 + 1 = 3\n"
        assert str(hello_py_doc.output()) == "3\n"

        assert hello_py_doc.final_artifact.content_source == 'generated'
        assert doc_txt_doc.final_artifact.content_source == 'generated'

        wrapper = Wrapper()

        hello_py = DocNode("hello.py|py", wrapper=wrapper)

        doc_txt = DocNode("doc.txt|jinja",
                inputs = [hello_py],
                wrapper=wrapper)

        wrapper.run_docs(doc_txt)

        hello_py_doc = hello_py.children[0]
        doc_txt_doc = doc_txt.children[0]

        assert hello_py_doc.final_artifact.content_source == 'cached'
        assert doc_txt_doc.final_artifact.content_source == 'cached'

        with open("doc.txt", "w") as f:
            f.write("1 + 1 = {{ d['hello.py|py'] }}\n")

        wrapper = Wrapper()

        hello_py = DocNode("hello.py|py", wrapper=wrapper)

        doc_txt = DocNode("doc.txt|jinja",
                inputs = [hello_py],
                wrapper=wrapper)

        wrapper.run_docs(doc_txt)

        hello_py_doc = hello_py.children[0]
        doc_txt_doc = doc_txt.children[0]

        assert hello_py_doc.final_artifact.content_source == 'cached'
        assert doc_txt_doc.final_artifact.content_source == 'generated'

        time.sleep(2)

        with open("hello.py", "w") as f:
            f.write("print 1+1\n")

        wrapper = Wrapper()

        hello_py = DocNode("hello.py|py", wrapper=wrapper)

        doc_txt = DocNode("doc.txt|jinja",
                inputs = [hello_py],
                wrapper=wrapper)

        wrapper.run_docs(doc_txt)

        hello_py_doc = hello_py.children[0]
        doc_txt_doc = doc_txt.children[0]

        assert hello_py_doc.final_artifact.content_source == 'generated'
        assert doc_txt_doc.final_artifact.content_source == 'generated'

def test_node_init():
    node = Node("foo.txt")
    assert node.key == "foo.txt"

def test_node_init_with_inputs():
    node = Node("foo.txt", inputs=[Node("bar.txt")])
    assert node.key == "foo.txt"
    assert node.inputs[0].key == "bar.txt"

    expected = {
            0 : "bar.txt",
            1 : "foo.txt" 
        }

    for i, n in enumerate(node.walk_inputs()):
        assert expected[i] == n.key

def test_doc_node_populate():
    with wrap() as wrapper:
        node = Task.create('doc', "foo.txt", wrapper=wrapper)
        node.populate()
        assert node.key_with_class() == "DocNode:foo.txt"
        assert node.children[0].key_with_class() == "Doc:foo.txt"

def test_doc_node_with_filters():
    with wrap() as wrapper:
        node = Task.create('doc', "foo.txt|outputabc", wrapper=wrapper)
        node.populate()
        assert node.key_with_class() == "DocNode:foo.txt|outputabc"
        assert node.children[0].key_with_class() == "Doc:foo.txt|outputabc"

def test_pattern_node():
    with wrap() as wrapper:
        with open("foo.txt", "w") as f:
            f.write("foo!")

        with open("bar.txt", "w") as f:
            f.write("bar!")

        wrapper.setup_batch()

        node = PatternNode("*.txt", foo="bar", wrapper=wrapper)
        node.populate()
        assert node.args['foo'] == 'bar'
        assert len(node.children) == 2

        for child in node.children:
            assert child.__class__.__name__ == "Doc"
            assert child.args['foo'] == 'bar'
            assert child.key_with_class() in ["Doc:foo.txt", "Doc:bar.txt"]
            assert child.filters == []

def test_pattern_node_multiple_filters():
    with wrap() as wrapper:
        with open("foo.txt", "w") as f:
            f.write("foo!")

        wrapper.setup_batch()

        node = PatternNode("*.txt|dexy|dexy|dexy", wrapper=wrapper)
        node.populate()
        doc = node.children[0]
        assert doc.key == "foo.txt|dexy|dexy|dexy"
        assert doc.filters == ['dexy', 'dexy', 'dexy']
        assert doc.node == node

def test_pattern_node_one_filter():
    with wrap() as wrapper:
        with open("foo.txt", "w") as f:
            f.write("foo!")

        wrapper.setup_batch()

        node = PatternNode("*.txt|dexy", wrapper=wrapper)
        node.populate()
        doc = node.children[0]
        assert doc.key == "foo.txt|dexy"
        assert doc.filters == ['dexy']
        assert doc.node == node

        assert doc.children[0].key_with_class() == "InitialArtifact:foo.txt(|dexy)"
        assert doc.children[1].key_with_class() == "FilterArtifact:foo.txt|dexy"

        assert doc.children[0].state == "populated"
        assert doc.children[1].state == "populated"



from dexy.exceptions import UserFeedback
from dexy.doc import Doc
from dexy.tests.utils import wrap
from nose.tools import raises
import os

def test_custom_name():
    with wrap() as wrapper:
        doc = Doc(
                "data.txt",
                canonical_name="data.abc",
                contents="12345.67",
                wrapper=wrapper
                )
        wrapper.run_docs(doc)

        assert doc.output().name == "data.abc"

def test_custom_name_in_subdir():
    with wrap() as wrapper:
        doc = Doc(
                "data.txt",
                canonical_name="subdir/data.abc",
                contents="12345.67",
                wrapper=wrapper
                )
        wrapper.run_docs(doc)
        wrapper.report()

        assert doc.output().name == "subdir/data.abc"
        assert doc.output().parent_dir() == "subdir"

def test_custom_name_with_args():
    with wrap() as wrapper:
        doc = Doc(
                "data.txt",
                canonical_name="%(bar)s/data-%(foo)s.abc",
                foo='bar',
                bar='baz',
                contents="12345.67",
                wrapper=wrapper
                )
        wrapper.run_docs(doc)
        wrapper.report()

        assert doc.output().name == "baz/data-bar.abc"
        assert doc.output().parent_dir() == "baz"
        assert os.path.exists("output/baz/data-bar.abc")

@raises(UserFeedback)
def test_custom_name_with_evil_args():
    with wrap() as wrapper:
        doc = Doc(
                "data.txt",
                canonical_name="/%(bar)s/data-%(foo)s.abc",
                foo='bar',
                bar='baz',
                contents="12345.67",
                wrapper=wrapper
                )
        wrapper.run_docs(doc)



from dexy.node import DocNode
from dexy.plugins.templating_filters import TemplateFilter
from dexy.plugins.templating_plugins import TemplatePlugin
from dexy.tests.utils import wrap
from nose.tools import raises
import dexy.exceptions

def test_jinja_indent_function():
    with wrap() as wrapper:
        node = DocNode("hello.txt|jinja",
                contents = """lines are:\n   {{ d['lines.txt'] | indent(3) }}""",
                inputs = [
                    DocNode("lines.txt",
                        contents = "line one\nline two",
                        wrapper=wrapper)
                    ],
                wrapper=wrapper)
        wrapper.run_docs(node)
        assert str(node.children[0].output()) == """lines are:
   line one
   line two"""

def test_jinja_kv():
    with wrap() as wrapper:
        node = DocNode("hello.txt|jinja",
                contents = """value of foo is '{{ d['blank.txt|keyvalueexample']['foo'] }}'""",
                inputs = [
                    DocNode("blank.txt|keyvalueexample",
                        contents = " ",
                        wrapper=wrapper)
                    ],
                wrapper=wrapper)
        wrapper.run_docs(node)
        doc = node.children[0]
        assert str(doc.output()) == "value of foo is 'bar'"

@raises(dexy.exceptions.UserFeedback)
def test_jinja_sectioned_invalid_section():
    with wrap() as wrapper:
        doc = DocNode("hello.txt|jinja",
                contents = """first line is '{{ d['lines.txt|lines']['3'] }}'""",
                inputs = [
                    DocNode("lines.txt|lines",
                        contents = "line one\nline two",
                        wrapper=wrapper)
                    ],
                wrapper=wrapper)
        wrapper.run_docs(doc)

def test_jinja_sectioned():
    with wrap() as wrapper:
        node = DocNode("hello.txt|jinja",
                contents = """first line is '{{ d['lines.txt|lines']['1'] }}'""",
                inputs = [
                    DocNode("lines.txt|lines",
                        contents = "line one\nline two",
                        wrapper=wrapper)
                    ],
                wrapper=wrapper)
        wrapper.run_docs(node)
        doc = node.children[0]
        assert str(doc.output()) == "first line is 'line one'"

def test_jinja_json_convert_to_dict():
    with wrap() as wrapper:
        node = DocNode("hello.txt|jinja",
                contents = """foo is {{ d['input.json'].json_as_dict()['foo'] }}""",
                inputs = [
                    DocNode("input.json",
                        contents = """{"foo":123}""",
                        wrapper=wrapper)
                    ],
                wrapper=wrapper)
        wrapper.run_docs(node)
        doc = node.children[0]
        assert str(doc.output()) == "foo is 123"

@raises(dexy.exceptions.UserFeedback)
def test_jinja_json():
    with wrap() as wrapper:
        node = DocNode("hello.txt|jinja",
                contents = """foo is {{ d['input.json']['foo'] }}""",
                inputs = [
                    DocNode("input.json",
                        contents = """{"foo":123}""",
                        wrapper=wrapper)
                    ],
                wrapper=wrapper)
        wrapper.run_docs(node)

@raises(dexy.exceptions.UserFeedback)
def test_jinja_undefined():
    with wrap() as wrapper:
        node = DocNode("template.txt|jinja",
                contents = """{{ foo }}""",
                wrapper=wrapper)

        wrapper.run_docs(node)

@raises(dexy.exceptions.UserFeedback)
def test_jinja_syntax_error():
    with wrap() as wrapper:
        node = DocNode("template.txt|jinja",
                contents = """{% < set foo = 'bar' -%}\nfoo is {{ foo }}\n""",
                wrapper=wrapper)

        wrapper.run_docs(node)

def test_jinja_filter_inputs():
    with wrap() as wrapper:
        node = DocNode("template.txt|jinja",
                contents = "The input is '{{ d['input.txt'] }}'",
                inputs = [
                    DocNode("input.txt",
                        contents = "I am the input.",
                        wrapper=wrapper)
                    ],
                wrapper=wrapper)

        wrapper.run_docs(node)
        doc = node.children[0]
        assert doc.output().as_text() == "The input is 'I am the input.'"

class TestSimple(TemplatePlugin):
    """
    test plugin
    """
    ALIASES = ['testtemplate']
    def run(self):
        return {'aaa' : 1}

class TestTemplateFilter(TemplateFilter):
    """
    test template
    """
    ALIASES = ['testtemplatefilter']

def test_template_filter_with_custom_filter_only():
    with wrap() as wrapper:
        node = DocNode("hello.txt|testtemplatefilter",
                contents = "aaa equals %(aaa)s",
                testtemplatefilter = { "plugins" : ["testtemplate"] },
                wrapper=wrapper)

        wrapper.run_docs(node)
        doc = node.children[0]
        assert doc.output().as_text() == "aaa equals 1"
        plugins_used = doc.final_artifact.filter_instance.template_plugins()
        assert len(plugins_used) == 1
        assert plugins_used[0] == TestSimple

def test_jinja_filter():
    with wrap() as wrapper:
        node = DocNode("template.txt|jinja",
                contents = "1 + 1 is {{ 1+1 }}",
                wrapper=wrapper)

        wrapper.run_docs(node)
        doc = node.children[0]
        assert doc.output().as_text() == "1 + 1 is 2"

def test_jinja_filter_tex_extension():
    with wrap() as wrapper:
        node = DocNode("template.tex|jinja",
                contents = "1 + 1 is << 1+1 >>",
                wrapper=wrapper)

        wrapper.run_docs(node)
        doc = node.children[0]
        assert doc.output().as_text() == "1 + 1 is 2"

def test_jinja_filter_custom_delims():
    with wrap() as wrapper:
        node = DocNode("template.tex|jinja",
                contents = "1 + 1 is %- 1+1 -%",
                jinja = {
                    "variable_start_string" : "%-",
                    "variable_end_string" : "-%"
                    },
                wrapper=wrapper)

        wrapper.run_docs(node)
        doc = node.children[0]
        assert doc.output().as_text() == "1 + 1 is 2"

def test_jinja_filter_set_vars():
    with wrap() as wrapper:
        node = DocNode("template.txt|jinja",
                contents = """{% set foo = 'bar' -%}\nfoo is {{ foo }}\n""",
                wrapper=wrapper)

        wrapper.run_docs(node)
        doc = node.children[0]
        assert doc.output().as_text() == "foo is bar"



from dexy.plugins.pexpect_filters import PexpectReplFilter
from dexy.plugins.process_filters import SubprocessCompileFilter
from dexy.plugins.process_filters import SubprocessStdoutFilter
import os
import platform

class JrubyFilter(SubprocessStdoutFilter):
    """
    Run jruby code and return stdout.
    """
    ALIASES = ['jruby']
    EXECUTABLE = "jruby"
    INPUT_EXTENSIONS = [".rb", ".txt"]
    OUTPUT_EXTENSIONS = [".txt"]
    VERSION_COMMAND = "jruby --version"

class JirbFilter(PexpectReplFilter):
    """
    Run jruby code in jirb.
    """
    ALIASES = ['jirb']
    ALLOW_MATCH_PROMPT_WITHOUT_NEWLINE = True
    CHECK_RETURN_CODE = False
    EXECUTABLE = "jirb --prompt-mode simple"
    INITIAL_PROMPT_TIMEOUT = 30
    INPUT_EXTENSIONS = [".rb", ".txt"]
    OUTPUT_EXTENSIONS = [".rbcon"]
    PROMPTS = ['>>', '?>']
    VERSION_COMMAND = "jirb --version"

class JythonFilter(SubprocessStdoutFilter):
    """
    jython
    """
    ALIASES = ['jython']
    EXECUTABLE = "jython"
    INPUT_EXTENSIONS = [".py", ".txt"]
    OUTPUT_EXTENSIONS = [".txt"]
    VERSION_COMMAND = "jython --version"

    @classmethod
    def is_active(klass):
        if platform.system() in ('Linux', 'Windows'):
            return klass.executable() and True or False
        elif platform.system() in ('Darwin'):
            if hasattr(klass, 'log'):
                klass.log.warn("The jython dexy filter should not be run on MacOS due to a serious bug. This filter is being disabled.")
            return False
        else:
            if hasattr(klass, 'log'):
                klass.log.warn("""Can't detect your system. If you see this message please report this to the dexy project maintainer, your platform.system() value is '%s'. The jython dexy filter should not be run on MacOS due to a serious bug.""" % platform.system())
            return klass.executable() and True or False

class JythonInteractiveFilter(PexpectReplFilter):
    """
    jython in REPL
    """
    ALIASES = ['jythoni']
    CHECK_RETURN_CODE = False
    EXECUTABLE = "jython -i"
    INITIAL_PROMPT_TIMEOUT = 30
    INPUT_EXTENSIONS = [".py", ".txt"]
    OUTPUT_EXTENSIONS = [".pycon"]
    VERSION_COMMAND = "jython --version"

    @classmethod
    def is_active(klass):
        if platform.system() in ('Linux', 'Windows'):
            return klass.executable() and True or False
        elif platform.system() in ('Darwin'):
            print "The jythoni dexy filter should not be run on MacOS due to a serious bug. This filter is being disabled."
            return False
        else:
            print """Can't detect your system. If you see this message please report this to the dexy project maintainer, your platform.system() value is '%s'. The jythoni dexy filter should not be run on MacOS due to a serious bug.""" % platform.system()
            return klass.executable() and True or False

class JavaFilter(SubprocessCompileFilter):
    """
    Compiles java code and runs main method.
    """
    ALIASES = ['java']
    CHECK_RETURN_CODE = True # Whether to check return code when running compiled executable.
    COMPILED_EXTENSION = ".class"
    EXECUTABLE = "javac"
    INPUT_EXTENSIONS = [".java"]
    OUTPUT_EXTENSIONS = [".txt"]
    VERSION_COMMAND = "java -version"

    def setup_cp(self):
        """
        Makes sure the current working directory is on the classpath, also adds
        any specified CLASSPATH elements. Assumes that CLASSPATH elements are either
        absolute paths, or paths relative to the artifacts directory. Also, if
        an input has been passed through the javac filter, its directory is
        added to the classpath.
        """
        self.log.debug("in setup_cp for %s" % self.artifact.key)

        classpath_elements = []

        working_dir = os.path.join(self.artifact.tmp_dir(), self.output().parent_dir())
        abs_working_dir = os.path.abspath(working_dir)
        self.log.debug("Adding working dir %s to classpath" % abs_working_dir)
        classpath_elements.append(abs_working_dir)

        for doc in self.processed():
            if (doc.output().ext == ".class") and ("javac" in doc.key):
                classpath_elements.append(doc.output().parent_dir())

        for item in self.args().get('classpath', []):
            for x in item.split(":"):
                classpath_elements.append(x)

        env = self.setup_env()
        if env and env.has_key("CLASSPATH"):
            for x in env['CLASSPATH'].split(":"):
                classpath_elements.append(x)

        cp = ":".join(classpath_elements)
        self.log.debug("Classpath %s" % cp)
        return cp

    def compile_command_string(self):
        cp = self.setup_cp()
        basename = os.path.basename(self.input().name)
        if len(cp) == 0:
            return "javac %s" % basename
        else:
            return "javac -classpath %s %s" % (cp, basename)

    def run_command_string(self):
        cp = self.setup_cp()
        main_method = self.setup_main_method()
        args = self.command_line_args() or ""
        return "java %s -cp %s %s" % (args, cp, main_method)

    def setup_main_method(self):
        basename = os.path.basename(self.input().name)
        default_main = os.path.splitext(basename)[0]
        return self.args().get('main', default_main)

class JavacFilter(JavaFilter):
    """
    Compiles java code and returns the .class object
    """
    ALIASES = ['javac']
    EXECUTABLE = "javac"
    INPUT_EXTENSIONS = [".java"]
    OUTPUT_EXTENSIONS = [".class"]
    VERSION_COMMAND = "java -version"

    def process(self):
        # Compile the code
        command = self.compile_command_string()
        proc, stdout = self.run_command(command, self.setup_env())
        self.handle_subprocess_proc_return(command, proc.returncode, stdout)
        self.copy_canonical_file()



from dexy.common import OrderedDict
from dexy.exceptions import InternalDexyProblem
from dexy.exceptions import UserFeedback
from dexy.plugins.process_filters import SubprocessFilter
import re

try:
    import pexpect
    AVAILABLE = True
except ImportError:
    AVAILABLE = False

class DexyEOFException(UserFeedback):
    pass

class PexpectReplFilter(SubprocessFilter):
    """
    Base class for filters which use pexpect to retrieve output line-by-line based on detecting prompts.
    """
    ALLOW_MATCH_PROMPT_WITHOUT_NEWLINE = False
    INITIAL_PROMPT = None
    INITIAL_PROMPT_TIMEOUT = 5
    TIMEOUT = 10 # Set to larger number if needed. Or set to None for no timeout.
    LINE_ENDING = "\n"
    OUTPUT_DATA_TYPE = 'sectioned'
    PROMPTS = ['>>>', '...'] # Python uses >>> prompt normally and ... when in multi-line structures like loops
    PROMPT_REGEX = None
    PS1 = None
    PS2 = None
    PS3 = None
    PS4 = None
    SAVE_VARS_TO_JSON_CMD = None
    TRIM_PROMPT = '>>>'
    STRIP_REGEX = None

    @classmethod
    def is_active(klass):
        return AVAILABLE and klass.executable() and klass.required_executables_present()

    def prompt_search_terms(self):
        """
        Search first for the prompt (or prompts) following a line ending.
        Also optionally allow matching the prompt with no preceding line ending.
        """

        if self.PROMPT_REGEX:
            prompts = [self.PROMPT_REGEX]
        elif hasattr(self, 'PROMPT'):
            prompts = [self.PROMPT]
        else:
            prompts = self.PROMPTS

        if self.ALLOW_MATCH_PROMPT_WITHOUT_NEWLINE:
            return ["%s%s" % (self.LINE_ENDING, p) for p in prompts] + prompts
        else:
            return ["%s%s" % (self.LINE_ENDING, p) for p in prompts]

    def lines_for_section(self, section_text):
        """
        Take the section text and split it into lines which will be sent to the
        T
        differently, or if you don't want the extra newline at the end.
        """
        return section_text.splitlines() + ["\n"]

    def strip_trailing_prompts(self, section_transcript):
        lines = section_transcript.splitlines()
        while len(lines) > 0 and re.match("^\s*(%s)\s*$|^\s*$" % self.TRIM_PROMPT, lines[-1]):
            lines = lines[0:-1]
        return self.LINE_ENDING.join(lines)

    def strip_newlines(self, line):
        return line.replace(" \r", "")

    def section_output(self, input_dict):
        """
        Runs the code in sections and returns an iterator so we can do custom stuff.
        """
        # If we want to automatically record values of local variables in the
        # script we are running, we add a section at the end of script
        do_record_vars = self.args().get('record_vars', False) or self.args().get('record-vars', False)
        if do_record_vars:
            if not self.SAVE_VARS_TO_JSON_CMD:
                raise UserFeedback("You specified record-vars but this option isn't available since SAVE_VARS_TO_JSON_CMD is not set for this filter.")

            section_text = self.SAVE_VARS_TO_JSON_CMD % self.input().basename()
            self.log.debug("Adding SAVE_VARS_TO_JSON_CMD code:\n%s" % section_text)
            input_dict['dexy--save-vars'] = section_text
            if not self.args().get('add-new-files'):
                self.args()['add-new-files'] = (".json")

        search_terms = self.prompt_search_terms()

        env = self.setup_env()

        if self.args().has_key('PS1'):
            ps1 = self.arg_value('PS1')
            self.log.debug("Setting PS1 to %s" % ps1)
            env['PS1'] = ps1
        elif self.PS1:
            self.log.debug("Setting PS1 to %s" % self.PS1)
            env['PS1'] = self.PS1

        if self.args().has_key('PS2'):
            ps2 = self.arg_value('PS2')
            self.log.debug("Setting PS2 to %s" % ps2)
            env['PS2'] = ps2
        elif self.PS2:
            self.log.debug("Setting PS2 to %s" % self.PS2)
            env['PS2'] = self.PS2

        if self.args().has_key('PS3'):
            ps3 = self.arg_value('PS3')
            self.log.debug("Setting PS3 to %s" % ps3)
            env['PS3'] = ps3
        elif self.PS3:
            self.log.debug("Setting PS3 to %s" % self.PS3)
            env['PS3'] = self.PS3

        if self.args().has_key('PS4'):
            ps4 = self.arg_value('PS4')
            self.log.debug("Setting PS4 to %s" % ps4)
            env['PS4'] = ps4
        elif self.PS4:
            self.log.debug("Setting PS4 to %s" % self.PS4)
            env['PS4'] = self.PS4

        timeout = self.setup_timeout()
        initial_timeout = self.setup_initial_timeout()

        wd=self.setup_wd()
        self.log.debug("About to spawn new process '%s' in %s." % (self.executable(), wd))

        # Spawn the process
        proc = pexpect.spawn(
                self.executable(),
                cwd=wd,
                env=env)

        self.log.debug("Capturing initial prompt...")
        try:
            if self.INITIAL_PROMPT:
                proc.expect(self.INITIAL_PROMPT, timeout=initial_timeout)
            elif self.PROMPT_REGEX:
                proc.expect(search_terms, timeout=initial_timeout)
            else:
                proc.expect_exact(search_terms, timeout=initial_timeout)

        except pexpect.TIMEOUT:
            if self.INITIAL_PROMPT:
                match = self.INITIAL_PROMPT
            else:
                match = search_terms

            msg = "%s failed at matching initial prompt within %s seconds. " % (self.__class__.__name__, initial_timeout)
            msg += "Received '%s', tried to match with '%s'" % (proc.before, match)
            msg += "\nExact characters received:\n"
            for i, c in enumerate(proc.before):
                msg += "chr %02d: %s\n" % (i, ord(c))
            msg += "The developer might need to set a longer INITIAL_PROMPT_TIMEOUT or the regexp may be wrong."
            raise InternalDexyProblem(msg)

        start = proc.before + proc.after

        self.log.debug(u"Initial prompt captured!")
        self.log.debug(unicode(start))

        for section_key, section_text in input_dict.items():
            section_transcript = start
            start = ""

            lines = self.lines_for_section(section_text)
            for l in lines:
                self.log.debug(u"Sending '%s'" % l)
                section_transcript += start
                proc.send(l.rstrip() + "\n")
                try:
                    if self.PROMPT_REGEX:
                        proc.expect(search_terms, timeout=timeout)
                    else:
                        proc.expect_exact(search_terms, timeout=timeout)


                    self.log.debug(u"Received '%s'" % unicode(proc.before, errors='replace'))

                    section_transcript += self.strip_newlines(proc.before)
                    start = proc.after
                except pexpect.EOF:
                    self.log.debug("EOF occurred!")
                    if not self.ignore_errors():
                        raise DexyEOFException()
                except pexpect.TIMEOUT:
                    msg = "Failed at matching prompt within %s seconds. " % timeout
                    msg += "Received '%s', tried to match with '%s'" % (proc.before, search_terms)
                    msg += "Something may have gone wrong, or you may need to set a longer timeout."
                    raise UserFeedback(msg)
                except pexpect.ExceptionPexpect as e:
                    raise UserFeedback(str(e))
                except pexpect.EOF as e:
                    raise UserFeedback(str(e))

            if self.STRIP_REGEX:
                section_transcript = re.sub(self.STRIP_REGEX, "", section_transcript)

            yield section_key, section_transcript

        if self.do_add_new_files():
            self.add_new_files()

        try:
            proc.close()
        except pexpect.ExceptionPexpect:
            raise UserFeedback("process %s may not have closed" % proc.pid)

        if proc.exitstatus and self.CHECK_RETURN_CODE:
            self.handle_subprocess_proc_return(self.executable(), proc.exitstatus, section_transcript)

    def process(self):
        output = OrderedDict()

        self.log.debug("args: " % self.args())
        for section_key, section_transcript in self.section_output(self.input().as_sectioned()):
            self.log.debug("Processing section %s" % section_key)
            section_output = self.strip_trailing_prompts(section_transcript)

            output[section_key] = section_output

        self.output().set_data(output)

class RubyPexpectReplFilter(PexpectReplFilter):
    """
    Runs ruby code in irb.
    """
    ALIASES = ['irb', 'rbrepl']
    EXECUTABLE = 'irb --simple-prompt'
    CHECK_RETURN_CODE = False
    INITIAL_PROMPT = "^>>"
    INPUT_EXTENSIONS = [".txt", ".rb"]
    OUTPUT_EXTENSIONS = [".rbcon"]
    OUTPUT_LEXER = "irb"
    PROMPTS = [">>", "?>"]
    TRIM_PROMPT = '>>'
    VERSION_COMMAND = 'irb --version'

class PythonPexpectReplFilter(PexpectReplFilter):
    """
    Runs python code in python's REPL.
    """
    ADD_NEW_FILES = True
    ALIASES = ['pycon', 'pyrepl']
    CHECK_RETURN_CODE = False
    EXECUTABLE = 'python'
    INPUT_EXTENSIONS = [".txt", ".py"]
    OUTPUT_EXTENSIONS = [".pycon"]
    OUTPUT_LEXER = "pycon"
    TAGS = ['python', 'interpreter', 'language']
    VERSION_COMMAND = 'python --version'

    SAVE_VARS_TO_JSON_CMD = """
import json
with open("%s-vars.json", "w") as dexy__vars_file:
    dexy__x = {}
    for dexy__k, dexy__v in locals().items():
        dexy__x[dexy__k] = str(dexy__v)
    json.dump(dexy__x, dexy__vars_file)
"""

try:
    import IPython
    IPYTHON_AVAILABLE = True
except ImportError:
    IPYTHON_AVAILABLE = False

class IpythonPexpectReplFilter(PythonPexpectReplFilter):
    """
    Runs python code in ipython.
    """
    # TODO test for version of ipython that supports --classic
    ALIASES = ['ipython']
    EXECUTABLE = 'ipython --classic'
    CHECK_RETURN_CODE = False # TODO try to figure out why we are getting nonzero exit codes
    INPUT_EXTENSIONS = [".txt", ".py"]
    OUTPUT_EXTENSIONS = [".pycon"]
    VERSION_COMMAND = 'ipython -Version'

    @classmethod
    def is_active(klass):
        return klass.executable() and IPYTHON_AVAILABLE

class RPexpectReplFilter(PexpectReplFilter):
    """
    Runs R in REPL.
    """
    ADD_NEW_FILES = True
    ALIASES = ['r', 'rint']
    CHECK_RETURN_CODE = False
    EXECUTABLE = "R --quiet --vanilla"
    INPUT_EXTENSIONS = ['.txt', '.r', '.R']
    OUTPUT_EXTENSIONS = ['.Rout']
    PROMPT_REGEX = "(\x1b[^m]*m)?(>|\+)\s*"
    INITIAL_PROMPT = "(\x1b[^>])?>\s*"
    TAGS = ['r', 'interpreter', 'language']
    TRIM_PROMPT = ">"
    STRIP_REGEX = "(\x1b[^h]+h)" # Strip weird initial prompt on OSX
    VERSION_COMMAND = "R --version"
    SAVE_VARS_TO_JSON_CMD = """
if ("rjson" %%in%% installed.packages()) {
    library(rjson)
    dexy__json_file <- file("%s", "w")
    writeLines(toJSON(as.list(environment())), dexy__json_file)
    close(dexy__json_file)
} else {
   cat("Can't automatically save environment to JSON since rjson package not installed.")
}
"""

class RhinoInteractiveFilter(PexpectReplFilter):
    """
    Runs rhino JavaScript interpeter.
    """
    EXECUTABLE = "rhino"
    INPUT_EXTENSIONS = [".js", ".txt"]
    OUTPUT_EXTENSIONS = [".jscon"]
    ALIASES = ['jsint', 'rhinoint']
    PROMPTS = ['js>', '  >']
    TRIM_PROMPT = "js>"
    INITIAL_PROMPT_TIMEOUT = 60

class PhpInteractiveFilter(PexpectReplFilter):
    """
    Runs PHP in interpeter mode.
    """
    CHECK_RETURN_CODE = False
    EXECUTABLE = "php -a"
    INPUT_EXTENSIONS = [".php", ".txt"]
    OUTPUT_EXTENSIONS = [".txt"]
    ALIASES = ['phpint']
    PROMPTS = ['php > ']
    TRIM_PROMPT = "php > "

class BashInteractiveFilter(PexpectReplFilter):
    """
    Runs bash. Use to run bash scripts.
    """
    ALIASES = ['shint', 'bashint']
    EXECUTABLE = "bash --norc -i"
    INPUT_EXTENSIONS = [".txt", ".sh"]
    OUTPUT_EXTENSIONS = ['.sh-session']
    OUTPUT_LEXER = "console"
    PROMPT_REGEX = r"\d*[#$]"
    INITIAL_PROMPT = PROMPT_REGEX
    TRIM_PROMPT = PROMPT_REGEX
    PS1 = "$ "

class KshInteractiveFilter(PexpectReplFilter):
    """
    Runs ksh. Use to run bash scripts.
    """
    ALIASES = ['kshint']
    EXECUTABLE = "ksh -i"
    INPUT_EXTENSIONS = [".txt", ".sh"]
    OUTPUT_EXTENSIONS = ['.sh-session']
    INITIAL_PROMPT = "^\s*\d*(#|\$)\s+"
    PROMPT_REGEX = "\d*(#|\$)"
    TRIM_PROMPT = r"\d*(\$|#)"
    PS1 = "$ "

class MatlabInteractiveFilter(PexpectReplFilter):
    """
    Runs matlab in REPL.
    """
    ALIASES = ['matlabint']
    EXECUTABLE = 'matlab -nodesktop -nosplash -nodisplay'
    INPUT_EXTENSIONS = ['.m', '.txt']
    OUTPUT_EXTENSIONS = ['.mout']
    LINE_ENDING = "\r\n"
    PROMPT = ">>"
    # TODO handle EOF errors if people have 'quit' in their script

class ClojureInteractiveFilter(PexpectReplFilter):
    """
    Runs clojure.
    """
    ALIASES = ['clj', 'cljint']
    CHECK_RETURN_CODE = False
    EXECUTABLES = ['clojure -r', 'clj -r']
    INITIAL_PROMPT_TIMEOUT = 15
    INPUT_EXTENSIONS = [".clj", ".txt"]
    OUTPUT_EXTENSIONS = [".txt"]
    PROMPT = "user=> "

    def lines_for_section(self, input_text):
        input_lines = []
        current_line = []
        in_indented_block = False
        for l in input_text.splitlines():
            if re.match("^\s+", l):
                in_indented_block = True
                current_line.append(l)
            else:
                if len(current_line) > 0:
                    input_lines.append("\n".join(current_line))
                if in_indented_block:
                    # we have reached the end of this indented block
                    in_indented_block = False
                current_line = [l]
        input_lines.append("\n".join(current_line))
        return input_lines



from dexy.utils import os_to_posix
import dexy.doc
import dexy.exceptions
import dexy.plugin
import dexy.utils
import os
import posixpath

class FilterException(Exception):
    pass

class Filter:
    __metaclass__ = dexy.plugin.PluginMeta

    ALIASES = ['dexy']
    FRAGMENT = True
    INPUT_EXTENSIONS = [".*"]
    NODOC = False
    OUTPUT_DATA_TYPE = 'generic'
    OUTPUT_EXTENSIONS = [".*"]
    PRESERVE_PRIOR_DATA_CLASS = False
    REQUIRE_OUTPUT = True
    TAGS = []

    @classmethod
    def templates(klass):
        """
        List of dexy templates which refer to this filter.
        """
        import dexy.template
        return [p for p in dexy.template.Template.plugins if any(a for a in klass.ALIASES if a in p.FILTERS_USED)]

    @classmethod
    def is_active(klass):
        return True

    @classmethod
    def inactive_because_missing(klass):
        if hasattr(klass, 'executables'):
            return klass.executables()
        elif hasattr(klass, 'IMPORTS'):
            return klass.IMPORTS

    def args(self):
        return self.artifact.filter_args()

    def arg_value(self, arg_name_hyphen, default=None):
        return dexy.utils.value_for_hyphenated_or_underscored_arg(self.args(), arg_name_hyphen, default)

    @classmethod
    def data_class_alias(klass, file_ext):
        return klass.OUTPUT_DATA_TYPE

    def do_add_new_files(self):
        return self.ADD_NEW_FILES or self.arg_value("add-new-files", False)

    def process(self):
        pass

    def calculate_canonical_name(self):
        name_without_ext = posixpath.splitext(self.artifact.key)[0]
        return "%s%s" % (name_without_ext, self.artifact.ext)

    def doc_arg(self, arg_name_hyphen, default=None):
        return self.artifact.doc.arg_value(arg_name_hyphen, default)

    def add_doc(self, doc_name, doc_contents=None):
        doc_name = os_to_posix(doc_name)
        if not posixpath.sep in doc_name:
            doc_name = posixpath.join(self.input().parent_dir(), doc_name)

        additional_doc_filters = self.arg_value('additional-doc-filters', {}) 
        self.log.debug("additional-doc-filters are %s" % additional_doc_filters)

        doc_ext = os.path.splitext(doc_name)[1]

        if isinstance(additional_doc_filters, dict):
            filters = additional_doc_filters.get(doc_ext, '')
        elif isinstance(additional_doc_filters, str) or isinstance(additional_doc_filters, unicode):
            filters = additional_doc_filters
        else:
            # TODO allow passing a list of tuples so input can be
            # used in more than 1 way
            raise Exception("not implemented")

        if len(filters) > 0:
            if self.arg_value('keep-originals', True):
                doc_key = doc_name
                doc = dexy.doc.Doc(doc_key, contents=doc_contents)
                self.artifact.add_doc(doc)

            doc_key = "%s|%s" % (doc_name, filters)
            doc = dexy.doc.Doc(doc_key, contents=doc_contents)
            self.artifact.add_doc(doc)

        else:
            doc_key = doc_name
            doc = dexy.doc.Doc(doc_key, contents=doc_contents)
            self.artifact.add_doc(doc)

        return doc

    def input(self):
        return self.artifact.input_data

    def inputs(self):
        return self.artifact.doc.node.walk_inputs()

    def input_filename(self):
        return self.artifact.input_filename()

    def output_filename(self):
        return self.artifact.output_filename()

    def output(self):
        return self.artifact.output_data

    def output_filepath(self):
        return self.output().storage.data_file()

    def processed(self):
        return self.artifact.doc.node.walk_input_docs()

    def final_ext(self):
        return self.artifact.doc.final_artifact.ext

    def setup_wd(self, populate=True):
        wd = self.artifact.working_dir()
        wd_exists = os.path.exists(wd)
        self.log.debug("setting up wd for %s. exists already: %s" % (self.artifact.key, wd))
        written_already = set()
        if not wd_exists:
            for doc, filename in self.artifact.setup_wd(self.input_filename()):
                wa = filename in written_already
                self.write_to_wd(wd, doc, filename, wa)
                written_already.add(filename)

        return wd

    def write_to_wd(self, wd, doc, filename, wa=False):
        try:
            doc.output().output_to_file(filename)
        except Exception as e:
            args = (e.__class__.__name__, wd, self.artifact.key, doc.key, filename)
            self.log.debug("%s error occurred whlie trying to populate working directory %s for %s with %s (%s)" % args)
            self.log.debug(str(e))

    def resolve_conflict(self, doc, conflict_docs):
        """
        Return true if the doc wins the conflict and should be written to the canonical name, false if not.
        """
        conflict_docs = [d for d in conflict_docs if not (('pyg' in d.key) or ('idio' in d.key))]
        conflict_docs.sort()
        if len(conflict_docs) == 0:
            return True
        else:
            return doc in conflict_docs and conflict_docs.index(doc) == 0

class DexyFilter(Filter):
    """
    Filter which implements some default behaviors.
    """
    ALIASES = ['dexy']

    @classmethod
    def data_class_alias(klass, file_ext):
        if hasattr(klass, 'process_dict'):
            return 'sectioned'
        elif hasattr(klass, 'process_text_to_dict'):
            return 'sectioned'
        else:
            return klass.OUTPUT_DATA_TYPE

    def process(self):
        if hasattr(self, "process_text_to_dict"):
            output = self.process_text_to_dict(unicode(self.input()))
            self.output().set_data(output)

        elif hasattr(self, "process_dict"):
            output = self.process_dict(self.input().as_sectioned())
            self.output().set_data(output)

        elif hasattr(self, "process_text"):
            output = self.process_text(unicode(self.input()))
            self.output().set_data(output)

        else:
            self.output().copy_from_file(self.input().storage.data_file())

class AliasFilter(DexyFilter):
    """
    Filter to be used when an Alias is specified. Should not change input.
    """
    PRESERVE_PRIOR_DATA_CLASS = True
    ALIASES = []

    def calculate_canonical_name(self):
        return self.artifact.prior.filter_instance.calculate_canonical_name()



from dexy.plugin import PluginMeta
import copy
import dexy.doc
import dexy.exceptions
import os
import posixpath
import pprint

class AbstractSyntaxTree():
    def __init__(self, wrapper=None):
        self.lookup_table = {}
        self.tree = []
        self.root_nodes_ordered = False
        self.wrapper = wrapper
        self.default_args = []

    def default_args_for_directory(self, path):
        default_kwargs = {}
        dir_path = posixpath.dirname(posixpath.abspath(path))

        for d, args in self.default_args:
            if posixpath.abspath(d) in dir_path:
                self.wrapper.log.debug("applying default args for dir %s of %s" % (d, args))
                default_kwargs.update(args)
                self.wrapper.log.debug("after updating: %s" % default_kwargs)

        return default_kwargs

    def standardize_key(self, key):
        return Parser.standardize_key(key)

    def add_task_info(self, task_key, **kwargs):
        """
        Adds kw args to kwarg dict in lookup table dict for this task
        """
        task_key = self.standardize_key(task_key)
        self.wrapper.log.debug("adding task info for '%s'" % task_key)

        if not task_key in self.tree:
            self.tree.append(task_key)

        if self.lookup_table.has_key(task_key):
            self.lookup_table[task_key].update(kwargs)
        else:
            self.lookup_table[task_key] = kwargs
            if not kwargs.has_key('children') or kwargs.has_key('inputs'):
                self.lookup_table[task_key]['inputs'] = []

        self.clean_tree()

    def add_dependency(self, task_key, input_task_key):
        """
        Adds input to list of inputs in lookup table dict for this task.
        """
        task_key = self.standardize_key(task_key)
        input_task_key = self.standardize_key(input_task_key)
        self.wrapper.log.debug("adding dependency of '%s' on '%s'" % (input_task_key, task_key))

        if task_key == input_task_key:
            return

        if not task_key in self.tree:
            self.tree.append(task_key)

        if self.lookup_table.has_key(task_key):
            self.lookup_table[task_key]['inputs'].append(input_task_key)
        else:
            self.lookup_table[task_key] = { 'inputs' : [input_task_key] }

        if not self.lookup_table.has_key(input_task_key):
            self.lookup_table[input_task_key] = { 'inputs' : [] }

        self.clean_tree()

    def clean_tree(self):
        """
        Removes tasks which are already represented as inputs.
        """
        all_inputs = self.all_inputs()

        # make copy since can't iterate and remove from same tree
        treecopy = copy.deepcopy(self.tree)

        for task in treecopy:
            if task in all_inputs:
                self.tree.remove(task)

    def all_inputs(self):
        """
        Returns a set of all task keys identified as inputs of some other element.
        """
        all_inputs = set()
        for kwargs in self.lookup_table.values():
            all_inputs.update(kwargs['inputs'])
        return all_inputs

    def task_kwargs(self, task_key):
        """
        Returns the dict of kw args for a task
        """
        args = self.lookup_table[task_key].copy()
        del args['inputs']
        return args

    def task_inputs(self, parent_key):
        """
        Returns the list of inputs for a atsk
        """
        return self.lookup_table[parent_key]['inputs']

    def debug(self, log=None):
        text = []
        text.append('tree:')
        for item in self.tree:
            text.append("  %s" % item)
        if self.root_nodes_ordered:
            text.append("root notes ordered.")
        text.append('lookup table:')
        for k, v in self.lookup_table.iteritems():
            pformat_v = pprint.pformat(v).splitlines()
            if len(pformat_v) == 0:
                raise Exception("no lines in pformat_v!")
            elif len(pformat_v) == 1:
                text.append("    %s: %s" % (k, pformat_v[0]))
            else:
                text.append("    %s:" % k)
                for line in pformat_v:
                    text.append("      %s" % line)

        if log:
            log.debug("\n".join(text))
        else:
            for line in text:
                print line

    def walk(self):
        created_tasks = {}
        root_nodes = []

        def create_dexy_task(key, *inputs, **kwargs):
            if not key in created_tasks:
                msg = "creating task '%s' with inputs '%s' with original kwargs '%s'"
                self.wrapper.log.debug(msg % (key, inputs, kwargs))
                alias, pattern = Parser.qualify_key(key)
                
                kwargs_with_defaults = self.default_args_for_directory(pattern)
                kwargs_with_defaults.update(kwargs)
                kwargs_with_defaults['inputs'] = inputs

                task = dexy.task.Task.create(alias, pattern, **kwargs_with_defaults)
                task.args_before_defaults = kwargs
                created_tasks[key] = task
            return created_tasks[key]

        def parse_item(key):
            inputs = self.task_inputs(key)
            kwargs = self.task_kwargs(key)
            kwargs['wrapper'] = self.wrapper
            if kwargs.get('inactive') or kwargs.get('disabled'):
                return
            if not kwargs.get('default', True):
                if self.wrapper.full:
                    pass
                elif self.wrapper.target and key.startswith(self.wrapper.target):
                    pass
                else:
                    return

            input_tasks = [parse_item(i) for i in inputs if i]

            # filter out inactive inputs
            input_tasks = [i for i in input_tasks if i]
    
            return create_dexy_task(key, *input_tasks, **kwargs)

        for key in self.tree:
            task = parse_item(key)
            if task:
                root_nodes.append(task)

        return root_nodes, created_tasks

class Parser:
    """
    Parse various types of config file.
    """
    ALIASES = []

    __metaclass__ = PluginMeta

    @classmethod
    def is_active(klass):
        return True

    @classmethod
    def qualify_key(klass, key):
        """
        Returns key split into pattern and alias, figuring out alias if not explict.
        """
        if not key:
            raise dexy.exceptions.InternalDexyProblem("Trying to call qualify_key with key of '%s'!" % key)

        if ":" in key:
            # split qualified key into alias & pattern
            alias, pattern = key.split(":")
        else:
            # this is an unqualified key, figure out its alias
            pattern = key

            # Allow '.ext' instead of '*.ext', shorter + easier for YAML
            if pattern.startswith(".") and not pattern.startswith("./"):
                if not os.path.exists(pattern):
                    pattern = "*%s" % pattern

            filepath = pattern.split("|")[0]
            if os.path.exists(filepath) and not os.path.isdir(filepath):
                alias = 'doc'
            elif (not "." in pattern) and (not "|" in pattern):
                alias = 'bundle'
            elif "*" in pattern:
                alias = 'pattern'
            else:
                alias = 'doc'

        alias = klass.standardize_alias(alias)
        return alias, pattern

    @classmethod
    def standardize_alias(klass, alias):
        return dexy.task.Task.aliases[alias].ALIASES[0]

    @classmethod
    def standardize_key(klass, key):
        """
        Only standardized keys should be used in the AST, so we don't create 2
        entries for what turns out to be the same task.
        """
        alias, pattern = klass.qualify_key(key)
        return "%s:%s" % (alias, pattern)

    def __init__(self, wrapper=None, ast=None):
        self.wrapper = wrapper
        self.ast = ast

    def parse(self, input_text, directory="."):
        """
        Method for testing, after this can call batch.run()
        """
        self.ast = AbstractSyntaxTree()
        self.ast.wrapper = self.wrapper
        self.build_ast(directory, input_text)

        self.wrapper.batch = dexy.batch.Batch(self.wrapper)
        self.wrapper.batch.load_ast(self.ast)

    def build_ast(self, directory, input_text):
        raise Exception("Implement in subclass.")

    def join_dir(self, directory, key):
        if directory == ".":
            return key
        else:
            starts_with_dot = key.startswith(".") and not key.startswith("./")
            does_not_exist = not os.path.exists(os.path.join(directory, key))
            if starts_with_dot and does_not_exist:
                key = "*%s" % key
            return posixpath.join(directory, key)



from dexy.artifact import FilterArtifact
from dexy.artifact import InitialArtifact
from dexy.artifact import InitialVirtualArtifact
from dexy.common import OrderedDict
from dexy.doc import Doc
from dexy.node import DocNode
from dexy.tests.utils import tempdir
from dexy.tests.utils import wrap
from dexy.wrapper import Wrapper
import dexy.exceptions
import time

def test_create_working_dir():
    with wrap() as wrapper:
        wrapper.setup(True)
        c1 = DocNode("data.txt", contents="12345.67", wrapper=wrapper)
        c2 = DocNode("mymod.py", contents="FOO='bar'", wrapper=wrapper)
        node = DocNode("example.py|py",
                inputs = [c1, c2],
                wrapper=wrapper,
                contents="""\
with open("data.txt", "r") as f:
    print f.read()

import mymod
print mymod.FOO

import os
print sorted(os.listdir(os.getcwd()))
""")

        wrapper.run_docs(node)
        doc = node.children[0]

        output = str(doc.output())
        assert "12345.67" in output
        assert 'bar' in output
        assert "['data.txt', 'example.py', 'mymod.py', 'mymod.pyc']" in output

def test_no_data():
    with wrap() as wrapper:
        node = DocNode("hello.txt", wrapper=wrapper)
        try:
            wrapper.run_docs(node)
            assert False, 'should raise UserFeedback'
        except dexy.exceptions.UserFeedback as e:
            assert "No contents found" in e.message

def test_caching():
    with tempdir():
        wrapper1 = Wrapper()
        wrapper1.setup_dexy_dirs()

        with open("abc.txt", "w") as f:
            f.write("these are the contents")

        node1 = DocNode("abc.txt|dexy", wrapper=wrapper1)
        wrapper1.run_docs(node1)

        doc1 = node1.children[0]

        assert isinstance(doc1.children[0], InitialArtifact)
        hashstring_0_1 = doc1.children[0].hashstring

        assert isinstance(doc1.children[1], FilterArtifact)
        hashstring_1_1 = doc1.children[1].hashstring

        wrapper2 = Wrapper()
        node2 = DocNode("abc.txt|dexy", wrapper=wrapper2)
        wrapper2.run_docs(node2)

        doc2 = node2.children[0]

        assert isinstance(doc2.children[0], InitialArtifact)
        hashstring_0_2 = doc2.children[0].hashstring

        assert isinstance(doc2.children[1], FilterArtifact)
        hashstring_1_2 = doc2.children[1].hashstring

        assert hashstring_0_1 == hashstring_0_2
        assert hashstring_1_1 == hashstring_1_2

def test_caching_virtual_file():
    with tempdir():
        wrapper1 = Wrapper()
        wrapper1.setup_dexy_dirs()

        node1 = DocNode("abc.txt|dexy",
                contents = "these are the contents",
                wrapper=wrapper1)
        wrapper1.run_docs(node1)

        doc1 = node1.children[0]

        assert isinstance(doc1.children[0], InitialVirtualArtifact)
        hashstring_0_1 = doc1.children[0].hashstring

        assert isinstance(doc1.children[1], FilterArtifact)
        hashstring_1_1 = doc1.children[1].hashstring

        wrapper2 = Wrapper()
        node2 = DocNode(
                "abc.txt|dexy",
                contents = "these are the contents",
                wrapper=wrapper2)
        wrapper2.run_docs(node2)

        doc2 = node2.children[0]

        assert isinstance(doc2.children[0], InitialVirtualArtifact)
        hashstring_0_2 = doc2.children[0].hashstring

        assert isinstance(doc2.children[1], FilterArtifact)
        hashstring_1_2 = doc2.children[1].hashstring

        assert hashstring_0_1 == hashstring_0_2
        assert hashstring_1_1 == hashstring_1_2

def test_virtual_artifact():
    with wrap() as wrapper:
        a = InitialVirtualArtifact("abc.txt",
                contents="these are the contents",
                wrapper=wrapper)

        a.name = "abc.txt"
        a.doc = Doc("abc.txt")
        a.setup()
        a.run()

        assert a.output_data.is_cached()
        assert a.output_data.data() == "these are the contents"

def test_initial_artifact_hash():
    with wrap() as wrapper:
        filename = "source.txt"

        with open(filename, "w") as f:
            f.write("hello this is some text")

        artifact = InitialArtifact(filename, wrapper=wrapper)
        artifact.name = filename
        artifact.doc = Doc(filename)
        artifact.setup()
        artifact.run()

        first_hashstring = artifact.hashstring

        time.sleep(1.1) # make sure mtime is at least 1 second different

        with open(filename, "w") as f:
            f.write("hello this is different text")

        artifact = InitialArtifact(filename, wrapper=wrapper)
        artifact.name = filename
        artifact.doc = Doc(filename)
        artifact.setup()
        artifact.run()

        second_hashstring = artifact.hashstring

        assert first_hashstring != second_hashstring

def test_parent_doc_hash():
    with tempdir():
        args = [["hello.txt|newdoc", { "contents" : "hello" }]]
        wrapper = Wrapper(*args)
        wrapper.setup(True)
        wrapper.run()

        node = wrapper.batch.tree[0]
        doc = node.children[0]
        hashstring = doc.final_artifact.hashstring

        wrapper = Wrapper(*args)
        wrapper.setup_db()
        wrapper.setup_log()
        wrapper.setup_batch()
        rows = wrapper.db.get_child_hashes_in_previous_batch(hashstring)
        assert len(rows) == 3

def test_parent_doc_hash_2():
    with tempdir():
        args = [["hello.txt|newdoc", { "contents" : "hello" }]]
        wrapper = Wrapper(*args)
        wrapper.setup_dexy_dirs()
        wrapper.run()

        for task in wrapper.batch.lookup_table.values():
            if task.__class__.__name__ == 'FilterArtifact':
                assert task.content_source == 'generated'

        wrapper = Wrapper(*args)
        wrapper.run()

        for task in wrapper.batch.lookup_table.values():
            if task.__class__.__name__ == 'FilterArtifact':
                assert task.content_source == 'cached'

def test_bad_file_extension_exception():
    with wrap() as wrapper:
        doc = DocNode("hello.abc|py",
                contents="hello",
                wrapper=wrapper)

        try:
            wrapper.run_docs(doc)
            assert False, "should not be here"
        except dexy.exceptions.UserFeedback as e:
            assert "Filter 'py' in 'hello.abc|py' can't handle file extension .abc" in e.message

def test_custom_file_extension():
    with wrap() as wrapper:
        node = DocNode("hello.py|pyg",
                contents="""print "hello, world" """,
                pyg = { "ext" : ".tex" },
                wrapper=wrapper)
        wrapper.run_docs(node)
        doc = node.children[0]
        assert "begin{Verbatim}" in str(doc.output())

def test_choose_extension_from_overlap():
    with wrap() as wrapper:
        doc = Doc("hello.py|pyg|forcelatex",
                contents="""print "hello, world" """,
                wrapper=wrapper)
        wrapper.run_docs(doc)
        assert "begin{Verbatim}" in str(doc.output())

def test_no_file_extension_overlap():
    with wrap() as wrapper:
        doc = Doc("hello.txt|forcetext|forcehtml",
                contents="hello",
                wrapper=wrapper)

        try:
            wrapper.run_docs(doc)
            assert False, "UserFeedback should be raised"
        except dexy.exceptions.UserFeedback as e:
            assert "Filter forcehtml can't go after filter forcetext, no file extensions in common." in e.message

def test_virtual_artifact_data_class_generic():
    with wrap() as wrapper:
        doc = Doc("virtual.txt",
                contents = "virtual",
                wrapper=wrapper)
        doc.populate()
        artifact = doc.children[0]
        assert artifact.__class__.__name__ == "InitialVirtualArtifact"
        assert artifact.data_class_alias() == 'generic'

def test_virtual_artifact_data_class_sectioned():
    with wrap() as wrapper:
        contents = OrderedDict()
        contents['foo'] = 'bar'
        doc = Doc("virtual.txt",
                contents=contents,
                wrapper=wrapper)
        doc.populate()
        artifact = doc.children[0]
        assert artifact.__class__.__name__ == "InitialVirtualArtifact"
        assert artifact.data_class_alias() == 'sectioned'



from dexy.utils import s
import dexy.batch
import dexy.database
import dexy.doc
import dexy.parser
import dexy.reporter
import logging
import logging.handlers
import os
import shutil

class Wrapper(object):
    """
    Class that assists in interacting with dexy, including running dexy.
    """
    DEFAULTS = {
        'artifacts_dir' : 'artifacts',
        'config_file' : 'dexy.conf',
        'danger' : False,
        'db_alias' : 'sqlite3',
        'db_file' : 'dexy.sqlite3',
        'disable_tests' : False,
        'dont_use_cache' : False,
        'dry_run' : False,
        'encoding' : 'utf-8',
        'exclude' : '.git, .svn, tmp, cache',
        'exclude_also' : '',
        'full' : False,
        'globals' : '',
        'hashfunction' : 'md5',
        'ignore_nonzero_exit' : False,
        'log_dir' : 'logs',
        'log_file' : 'dexy.log',
        'log_format' : "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        'log_level' : "DEBUG",
        'profile' : False,
        'recurse' : True,
        'reports' : '',
        'siblings' : False,
        'silent' : False,
        'target' : False,
        'uselocals' : False
    }
    LOG_LEVELS = {
        'DEBUG' : logging.DEBUG,
        'INFO' : logging.INFO,
        'WARN' : logging.WARN
    }

    def __init__(self, *args, **kwargs):
        self.args = args
        self.initialize_attribute_defaults()
        self.update_attributes_from_kwargs(kwargs)

    def initialize_attribute_defaults(self):
        for name, value in self.DEFAULTS.iteritems():
            setattr(self, name, value)

    def update_attributes_from_kwargs(self, kwargs):
        for key, value in kwargs.iteritems():
            if not key in self.DEFAULTS:
                raise Exception("invalid kwargs %s" % key)
            setattr(self, key, value)

    def setup(self, setup_dirs=False, log_config=True):
        if setup_dirs:
            self.setup_dexy_dirs()
        self.check_dexy_dirs()
        self.setup_log()
        self.setup_db()
        if log_config:
            self.log_dexy_config()

    def run(self):
        self.setup()

        self.batch = self.init_batch()
        self.batch.run(self.target)

        self.save_db()

    def log_dexy_config(self):
        self.log.debug("dexy has config:")
        for k in sorted(self.__dict__):
            if not k in ('ast', 'args', 'db', 'log', 'tasks'):
                self.log.debug("  %s: %s" % (k, self.__dict__[k]))

    def db_path(self):
        return os.path.join(self.artifacts_dir, self.db_file)

    def log_path(self):
        return os.path.join(self.log_dir, self.log_file)

    def setup_batch(self):
        """
        Shortcut method for calling init_batch and assigning to batch instance variable.
        """
        self.batch = self.init_batch()

    def init_batch(self):
        batch = dexy.batch.Batch(self)

        if len(self.args) > 0:
            batch.tree = self.docs_from_args()
            batch.create_lookup_table()
        else:
            ast = self.load_doc_config()
            batch.load_ast(ast)

        return batch

    def run_docs(self, *docs):
        self.args = docs
        self.run()

    def setup_read(self, batch_id=None):
        self.setup(log_config=False)

        if batch_id:
            self.batch_id = batch_id
        else:
            self.batch_id = self.db.max_batch_id()

    def check_dexy_dirs(self):
        if not (os.path.exists(self.artifacts_dir) and os.path.exists(self.log_dir)):
            raise dexy.exceptions.UserFeedback("You need to run 'dexy setup' in this directory first.")

    def setup_dexy_dirs(self):
        if not os.path.exists(self.artifacts_dir):
            os.mkdir(self.artifacts_dir)
        if not os.path.exists(self.log_dir):
            os.mkdir(self.log_dir)

    def remove_dexy_dirs(self, reports=False):
        if os.path.exists(self.artifacts_dir):
            shutil.rmtree(self.artifacts_dir)
        if os.path.exists(self.log_dir):
            shutil.rmtree(self.log_dir)

        if reports:
            if isinstance(reports, bool):
                reports=dexy.reporter.Reporter.plugins

            for report in reports:
                report.remove_reports_dir()

    def logging_log_level(self):
        try:
            return self.LOG_LEVELS[self.log_level.upper()]
        except KeyError:
            msg = "'%s' is not a valid log level, check python logging module docs."
            raise dexy.exceptions.UserFeedback(msg % self.log_level)

    def setup_log(self):
        if not hasattr(self, 'log') or not self.log:
            self.log = logging.getLogger('dexy')
            self.log.setLevel(self.logging_log_level())

            handler = logging.handlers.RotatingFileHandler(
                    self.log_path(),
                    encoding="utf-8")

            formatter = logging.Formatter(self.log_format)
            handler.setFormatter(formatter)

            self.log.addHandler(handler)

    def setup_db(self):
        db_class = dexy.database.Database.aliases[self.db_alias]
        self.db = db_class(self)

    def docs_from_args(self):
        """
        Creates document objects from argument strings, returns array of newly created docs.
        """
        docs = []
        for arg in self.args:
            self.log.debug("Processing arg %s" % arg)
            doc = self.create_doc_from_arg(arg)
            if not doc:
                raise Exception("no doc created for %s" % arg)
            doc.wrapper = self
            docs.append(doc)
        return docs

    def create_doc_from_arg(self, arg, **kwargs):
        if isinstance(arg, dexy.task.Task):
            return arg

        elif isinstance(arg, list):
            if not isinstance(arg[0], basestring):
                msg = "First arg in %s should be a string" % arg
                raise dexy.exceptions.UserFeedback(msg)

            if not isinstance(arg[1], dict):
                msg = "Second arg in %s should be a dict" % arg
                raise dexy.exceptions.UserFeedback(msg)

            if kwargs:
                raise Exception("Shouldn't have kwargs if arg is a list")

            alias, pattern = dexy.parser.Parser.qualify_key(arg[0])
            return dexy.task.Task.create(alias, pattern, **arg[1])

        elif isinstance(arg, basestring):
            alias, pattern = dexy.parser.Parser.qualify_key(arg[0])
            return dexy.task.Task.create(alias, pattern, **kwargs)

        else:
            raise Exception("unknown arg type %s for arg %s" % (arg.__class__.__name__, arg))

    def save_db(self):
        self.db.save()

    def reports_dirs(self):
        return [c.REPORTS_DIR for c in dexy.reporter.Reporter.plugins]

    def report(self):
        if self.reports:
            self.log.debug("generating user-specified reports '%s'" % self.reports)
            reporters = []
            for alias in self.reports.split():
                reporter_class = dexy.reporter.Reporter.aliases[alias]
                self.log.debug("initializing reporter %s for alias %s" % (reporter_class.__name__, alias))
                reporters.append(reporter_class())
        else:
            self.log.debug("no reports specified, generating all reports for which ALLREPORTS is True")
            reporters = [c() for c in dexy.reporter.Reporter.plugins if c.ALLREPORTS]

        for reporter in reporters:
            self.log.debug("running reporter %s" % reporter.ALIASES[0])
            reporter.run(self)

    def is_valid_dexy_dir(self, dirpath, dirnames):
        nodexy_file = os.path.join(dirpath, '.nodexy')
        pip_delete_this_dir_file = os.path.join(dirpath, "pip-delete-this-directory.txt")
        if os.path.exists(nodexy_file):
            self.log.debug("  skipping directory '%s' and its children because .nodexy file found" % dirpath)
            dirnames[:] = []
            return False
        else:
            if os.path.exists(pip_delete_this_dir_file):
                print s("""WARNING pip left an old build/ file lying around!
                You probably want to cancel this dexy run (ctrl+c) and remove this directory first!
                Dexy will continue running unless you stop it...""")

            for x in self.exclude_dirs():
                if x in dirnames:
                    skipping_dir = os.path.join(dirpath, x)
                    self.log.debug("  skipping directory '%s' because it matches exclude '%s'" % (skipping_dir, x))
                    dirnames.remove(x)

            return True

    def load_doc_config(self):
        """
        Look for document config files in current working tree and load them.
        """
        ast = dexy.parser.AbstractSyntaxTree(self)

        config_files_used = []
        for dirpath, dirnames, filenames in os.walk("."):
            if self.is_valid_dexy_dir(dirpath, dirnames):
                check_for_double_config = []
                for alias in dexy.parser.Parser.aliases.keys():
                    path_to_config = os.path.join(dirpath, alias)

                    if os.path.exists(path_to_config):
                        self.log.debug("loading config from '%s'" % path_to_config)
                        with open(path_to_config, "r") as f:
                            config_text = f.read()

                        check_for_double_config.append(path_to_config)
                        config_files_used.append(path_to_config)

                        parser = dexy.parser.Parser.aliases[alias](self, ast)
                        parser.build_ast(dirpath, config_text)

                if len(check_for_double_config) > 1:
                    msg = "more than one config file found in dir %s: %s"
                    msg_args = (dirpath, ", ".join(check_for_double_config))
                    raise dexy.exceptions.UserFeedback(msg % msg_args)

        if len(config_files_used) == 0:
            msg = "WARNING: Didn't find any document config files (like %s)"
            print msg % (", ".join(dexy.parser.Parser.aliases.keys()))
        self.log.debug("AST completed:")
        ast.debug(self.log)

        return ast

    def setup_config(self):
        self.setup_dexy_dirs()
        self.setup_log()
        self.load_doc_config()

    def cleanup_partial_run(self):
        if hasattr(self, 'db'):
            # TODO remove any entries which don't have
            self.db.save()

    def exclude_dirs(self):
        exclude_str = self.exclude
        if self.exclude_also:
            exclude_str += ",%s" % self.exclude_also
        exclude = [d.strip() for d in exclude_str.split(",")]

        exclude.append(self.artifacts_dir)
        exclude.append(self.log_dir)
        exclude.extend(self.reports_dirs())

        return exclude

    def parse_globals(self):
        globals_dict = {}
        if len(self.globals) > 0:
            for pair in self.globals.split(","):
                x, y = pair.split("=")
                globals_dict[x] = y

        return globals_dict

    def walk(self, start, recurse=True):
        exclude = self.exclude_dirs()

        for dirpath, dirnames, filenames in os.walk(start):
            for x in exclude:
                if x in dirnames:
                    dirnames.remove(x)

            if not recurse:
                dirnames[:] = []

            nodexy_file = os.path.join(dirpath, '.nodexy')
            if os.path.exists(nodexy_file):
                dirnames[:] = []
            else:
                for filename in filenames:
                    yield(dirpath, filename)



from setuptools import setup, find_packages
from dexy.version import DEXY_VERSION

setup(
        author='Ana Nelson',
        author_email='ana@ananelson.com',
        classifiers=[
            "Development Status :: 4 - Beta",
            "Environment :: Console",
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Intended Audience :: Financial and Insurance Industry",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: MIT License",
            "Topic :: Documentation",
            "Topic :: Software Development :: Build Tools",
            "Topic :: Software Development :: Code Generators",
            "Topic :: Software Development :: Documentation",
            "Topic :: Text Processing",
            "Topic :: Text Processing :: Markup :: HTML",
            "Topic :: Text Processing :: Markup :: LaTeX"
            ],
        dependency_links = [ "http://dexy.it/external-dependencies/" ],
        description='Document Automation',
        entry_points = {
            'pygments.lexers' : [
                'rst+django = dexy.plugins.pygments_plugins:RstDjangoLexer'
                ],
            'console_scripts' : [
                'dexy = dexy.commands:run'
                ]
            },
        include_package_data = True,
        install_requires = [
            'PyYAML',
            'chardet',
            'idiopidae>=0.5.4',
            'jinja2',
            'mock',
            'nose',
            'ordereddict',
            'pexpect',
            'pygments',
            'python-modargs>=1.6',
            'requests>=0.10.6',
            'zapps>=0.5.1'
            ],
        name='dexy',
        packages=find_packages(),
        url='http://dexy.it',
        version=DEXY_VERSION
        )




from dexy.plugins.process_filters import SubprocessFilter
import codecs
import dexy.exceptions
import dexy.utils
import os
import subprocess

class LatexFilter(SubprocessFilter):
    """
    Generates a PDF file from LaTeX source.
    """
    INPUT_EXTENSIONS = [".tex", ".txt"]
    OUTPUT_EXTENSIONS = [".pdf", ".png"]
    EXECUTABLES = ['pdflatex', 'latex']
    ALIASES = ['latex']
    FRAGMENT = False

    def command_string(self):
        clargs = self.command_line_args() or ''

        self.log.debug("Command line args specified by user '%s'" % clargs)

        if "-interaction=" in clargs:
            interaction_args = ''
        else:
            interaction_args = '-interaction=batchmode'

        args = {
            'prog' : self.executable(),
            'inter' : interaction_args,
            'args' : clargs,
            'input_file' : self.input().basename()
        }
        return "%(prog)s %(inter)s %(args)s %(input_file)s" % args

    def process(self):
        wd = self.setup_wd()
        env = self.setup_env()

        latex_command = self.command_string()

        bibtex_command = None
        if dexy.utils.command_exists("bibtex"):
            bibtex_command = "bibtex %s" % os.path.splitext(self.output().basename())[0]

        def run_cmd(command):
            self.log.info("running %s in %s" % (command, wd))
            proc = subprocess.Popen(command, shell=True,
                                    cwd=wd,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    env=env)

            stdout, stderr = proc.communicate()
            self.log.debug(stdout)

        if bibtex_command:
            run_cmd(latex_command) #generate aux
            run_cmd(bibtex_command) #generate bbl

        run_cmd(latex_command) # first run
        run_cmd(latex_command) # second run - fix references
        run_cmd(latex_command) # third run - just to be sure

        if not os.path.exists(os.path.join(wd, self.output().basename())):
            msg = "Latex file not generated. Look for information in latex log in %s directory." % wd
            raise dexy.exceptions.UserFeedback(msg)

        if self.do_add_new_files():
            self.log.debug("adding new files found in %s for %s" % (self.artifact.tmp_dir(), self.artifact.key))
            self.add_new_files()

        self.copy_canonical_file()

class XelatexFilter(LatexFilter):
    """
    Runs .tex files using xelatex.
    """
    ALIASES = ['xetex', 'xelatex']
    EXECUTABLE = 'xelatex'

class TikzPgfFilter(LatexFilter):
    """
    Takes a snippet of Tikz code, wraps it in a LaTeX document, and renders it to PDF.
    """
    ALIASES = ['tikz']

    def process(self):
        latex_filename = self.output().basename().replace(self.artifact.ext, ".tex")

        # TODO allow setting tikz libraries per-document, or just include all of them?
        # TODO how to create a page size that just includes the content
        latex_header = """\documentclass[tikz]{standalone}
\usetikzlibrary{shapes.multipart}
\\begin{document}
        """
        latex_footer = "\n\end{document}"

        wd = self.setup_wd()

        work_path = os.path.join(wd, latex_filename)
        self.log.debug("writing latex header + tikz content to %s" % work_path)
        with codecs.open(work_path, "w", encoding="utf-8") as f:
            f.write(latex_header)
            f.write(self.input().as_text())
            f.write(latex_footer)

        latex_command = "%s -interaction=batchmode %s" % (self.executable(), latex_filename)

        def run_cmd(command):
            self.log.info("about to run %s in %s" % (command, wd))
            proc = subprocess.Popen(command, shell=True,
                                    cwd=wd,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.STDOUT,
                                    env=self.setup_env())

            stdout, stderr = proc.communicate()

            if proc.returncode > 2: # Set at 2 for now as this is highest I've hit, better to detect whether PDF has been generated?
                raise dexy.exceptions.UserFeedback("latex error, look for information in %s" %
                                latex_filename.replace(".tex", ".log"))
            elif proc.returncode > 0:
                self.log.warn("""A non-critical latex error has occurred running %s,
                status code returned was %s, look for information in %s""" % (
                self.artifact.key, proc.returncode,
                latex_filename.replace(".tex", ".log")))

        run_cmd(latex_command)

        self.copy_canonical_file()



import dexy.artifact
import dexy.exceptions
import dexy.filter
import dexy.task
import os
import posixpath

class Doc(dexy.task.Task):
    """
    Task subclass representing Documents.
    """
    ALIASES = ['document']

    def arg_value(self, arg_name_hyphen, default=None):
        return dexy.utils.value_for_hyphenated_or_underscored_arg(self.args, arg_name_hyphen, default)

    def filter_class_for_alias(self, alias):
        if alias == '':
            blank_alias_msg = "You have a trailing | or you have 2 | symbols together in your specification for %s"
            raise dexy.exceptions.UserFeedback(blank_alias_msg % self.key)
        elif alias.startswith("-"):
            filter_class = dexy.filter.AliasFilter
        else:
            try:
                filter_class = dexy.filter.Filter.aliases[alias]
            except KeyError:
                msg = "Dexy doesn't have a filter '%s' available." % alias

                all_plugins = dexy.filter.Filter.aliases.values()
                num_plugins = len(all_plugins)
                if num_plugins < 10:
                    plugin_list = ", ".join(p.__name__ for p in all_plugins)
                    msg += " Note that only %s plugins are available: %s" % (num_plugins, plugin_list)
                    msg += " There may be a problem loading plugins, adding 'import dexy.plugins' might help."

                raise dexy.exceptions.UserFeedback(msg)

        if not filter_class.is_active():
            raise dexy.exceptions.InactiveFilter(alias, self.key)

        return filter_class

    def names_to_docs(self):
        """
        Returns a dict whose keys are canonical names, whose values are lists
        of the docs that generate that name as their canonical output name.
        """
        names_to_docs = {}
        for doc in self.node.walk_input_docs():
            doc_name = doc.output().name
            if names_to_docs.has_key(doc_name):
                names_to_docs[doc_name].append(doc)
            else:
                names_to_docs[doc_name] = [doc]
        return names_to_docs

    def conflicts(self):
        """
        List of inputs to document where more than 1 doc generates same
        canonical filename.
        """
        conflicts = {}
        for k, v in self.names_to_docs().iteritems():
            if len(v) > 1:
                conflicts[k] = v
        return conflicts

    def is_index_page(self):
        fn = self.output().name
        # TODO index.json only if htmlsections in doc key..
        return fn.endswith("index.html") or fn.endswith("index.json")

    def title(self):
        if self.args.get('title'):
            return self.args.get('title')
        elif self.is_index_page():
            # use subdirectory we're in
            return posixpath.split(posixpath.dirname(self.name))[-1].capitalize()
        else:
            return self.name

    def output(self):
        """
        Returns a reference to the output_data Data object generated by the final filter.
        """
        final_state = self.final_artifact.state
        if not final_state == 'complete':
            if not final_state == 'setup' and len(self.filters) == 0:
                raise dexy.exceptions.InternalDexyProblem("Final artifact state is '%s'" % self.final_artifact.state)

        return self.final_artifact.output_data

    def add_artifact(self, artifact):
        self.children.append(artifact)
        self.final_artifact = artifact

    def setup_initial_artifact(self):
        if os.path.exists(self.name):
            initial = dexy.artifact.InitialArtifact(self.name, wrapper=self.wrapper)
        else:
            initial = dexy.artifact.InitialVirtualArtifact(self.name, wrapper=self.wrapper)

        initial.args = self.args
        initial.name = self.name
        initial.prior = None
        initial.doc = self
        initial.created_by_doc = self.created_by_doc
        initial.remaining_doc_filters = self.filters

        initial.transition('populated')
        self.add_artifact(initial)

    def setup_filter_artifact(self, key, filters):
        filter_alias = filters[-1]

        remaining_filters = self.filters[len(filters):len(self.filters)]
        is_last_filter = len(remaining_filters) == 0

        artifact = dexy.artifact.FilterArtifact(key, wrapper=self.wrapper)

        artifact.remaining_doc_filters = remaining_filters
        artifact.set_log()
        artifact.log.addHandler(self.log.handlers[0])

        # skip args that are only relevant to the doc or to the initial artifact
        skip_args = ['contents', 'contentshash', 'data-class-alias', 'depends']
        artifact.args = dict((k, v) for k, v in self.args.iteritems() if not k in skip_args)

        artifact.doc = self
        artifact.prior = self.children[-1]
        artifact.created_by_doc = self.created_by_doc

        artifact.filter_alias = filter_alias
        artifact.filter_class = self.filter_class_for_alias(filter_alias)
        artifact.setup_filter_instance()

        if not is_last_filter:
            next_filter_alias = self.filters[len(filters)]
            artifact.next_filter_alias = next_filter_alias
            artifact.next_filter_class = self.filter_class_for_alias(next_filter_alias)
            artifact.next_filter_name = artifact.next_filter_class.__name__
        else:
            artifact.next_filter_alias = None
            artifact.next_filter_class = None
            artifact.next_filter_name = None

        artifact.transition('populated')
        self.add_artifact(artifact)

    def setup(self):
        self.metadata.final_artifact_hashstring = self.final_artifact.hashstring
        self.set_hashstring()

    def populate(self):
        self.set_log()
        self.name = self.key.split("|")[0]
        self.filters = self.key.split("|")[1:]
        self.canon = self.args.get('canon', len(self.filters) == 0)

        self.setup_initial_artifact()

        for i in range(0,len(self.filters)):
            filters = self.filters[0:i+1]
            key = "%s|%s" % (self.name, "|".join(filters))
            self.setup_filter_artifact(key, filters)
            self.canon = self.canon or (not self.final_artifact.filter_class.FRAGMENT)




from dexy.plugins.parsers import OriginalDexyParser
from dexy.plugins.parsers import TextFileParser
from dexy.plugins.parsers import YamlFileParser
from dexy.tests.utils import wrap
import dexy.exceptions
import os

YAML_WITH_INACTIVE = """
foo:
    - inactive: True
"""

def test_parse_inactive():
    with wrap() as wrapper:
        parser = YamlFileParser()
        parser.wrapper = wrapper
        parser.parse(YAML_WITH_INACTIVE)

        wrapper.batch.run()
        assert len(wrapper.batch.tasks()) == 0

YAML_WITH_DEFAULT_OFF = """
foo:
    - default: False
"""

def test_parse_default():
    with wrap() as wrapper:
        parser = YamlFileParser()
        parser.wrapper = wrapper
        parser.parse(YAML_WITH_DEFAULT_OFF)

        wrapper.batch.run()
        assert len(wrapper.batch.tasks()) == 0

    with wrap() as wrapper:
        wrapper.full = True
        parser = YamlFileParser()
        parser.wrapper = wrapper
        parser.parse(YAML_WITH_DEFAULT_OFF)

        wrapper.batch.run()
        assert len(wrapper.batch.tasks()) == 1

def test_yaml_with_defaults():
    with wrap() as wrapper:
        os.makedirs("s1/s2")

        with open("s1/s2/hello.txt", "w") as f:
            f.write("hello")
        parser = YamlFileParser()
        parser.wrapper = wrapper
        parser.parse(YAML_WITH_DEFAULTS)

        assert wrapper.batch.tree[0].args['foo'] == 'bar'
    
def test_invalid_yaml():
    with wrap() as wrapper:
        parser = YamlFileParser()
        parser.wrapper = wrapper
        try:
            parser.parse(INVALID_YAML)
            assert False, "should raise UserFeedback"
        except dexy.exceptions.UserFeedback as e:
            assert 'YAML' in e.message

def test_yaml_parser():
    with wrap() as wrapper:
        parser = YamlFileParser()
        parser.wrapper = wrapper
        parser.parse(YAML)
        docs = wrapper.batch.tree
        for doc in docs:
            assert doc.__class__.__name__ == 'BundleNode'
            assert doc.key in ['code', 'wordpress']

def test_text_parser_blank_lines():
    with wrap() as wrapper:
        parser = TextFileParser()
        parser.wrapper = wrapper
        parser.parse("\n\n")
        docs = wrapper.batch.tree
        assert len(docs) == 0

def test_text_parser_comments():
    with wrap() as wrapper:
        parser = TextFileParser()
        parser.wrapper = wrapper
        parser.parse("""
        valid.doc
        # commented-out.doc
        """)

        docs = wrapper.batch.tree
        assert len(docs) == 1
        assert docs[0].key == "valid.doc"

def test_text_parser_valid_json():
    with wrap() as wrapper:
        parser = TextFileParser()
        parser.wrapper=wrapper
        parser.parse("""
        doc.txt { "contents" : 123 }
        """)

        docs = wrapper.batch.tree
        assert docs[0].key == "doc.txt"
        assert docs[0].args['contents'] == 123

def test_text_parser_invalid_json():
    with wrap() as wrapper:
        parser = TextFileParser()
        parser.wrapper = wrapper

        try:
            parser.parse("""
            doc.txt { "contents" : 123
            """)
            assert False, 'should raise UserFeedback'
        except dexy.exceptions.UserFeedback as e:
            assert 'unable to parse' in e.message

def test_text_parser():
    with wrap() as wrapper:
        with open("f1.py", "w") as f:
            f.write("print 'hello'")

        with open("f2.py", "w") as f:
            f.write("print 'hello'")

        with open("index.md", "w") as f:
            f.write("")

        parser = TextFileParser()
        parser.wrapper = wrapper
        parser.parse("""
        *.py
        *.py|pyg
        *.md|jinja
        """)

        wrapper.batch.run()

        docs = wrapper.batch.docs()
        assert len(docs) == 5

def test_text_parser_virtual_file():
    with wrap() as wrapper:
        parser = TextFileParser()
        parser.wrapper = wrapper
        parser.parse("""
        virtual.txt { "contents" : "hello" }
        """)

        wrapper.batch.run()
        docs = wrapper.batch.tree

        assert docs[0].key == "virtual.txt"
        assert str(docs[0].children[0].output()) == "hello"

def test_original_parser():
    with wrap() as wrapper:
        conf = """{
        "*.txt" : {}
        }"""

        parser = OriginalDexyParser()
        parser.wrapper = wrapper
        parser.parse(conf)

        assert wrapper.batch.tree[0].key_with_class() == "PatternNode:*.txt"

def test_original_parser_allinputs():
    with wrap() as wrapper:
        conf = """{
        "*.txt" : {},
        "hello.txt" : { "contents" : "Hello!" },
        "*.md|jinja" : { "allinputs" : true }
        }"""

        parser = OriginalDexyParser(wrapper)
        parser.wrapper = wrapper
        parser.parse(conf)

        assert len(wrapper.batch.tree) == 1
        assert wrapper.batch.tree[0].key_with_class() == "PatternNode:*.md|jinja"

INVALID_YAML = """
code:
    - abc
    def
"""

YAML = """
code:
    - .R|pyg:
         "pyg" : { "foo" : "bar" }
    - .R|idio

wordpress:
    - code
    - test.txt|jinja
    - .md|jinja|markdown|wp
"""

YAML_WITH_DEFAULTS = """
defaults:
    pyg: { lexer : moin }
    foo: bar

code:
    - .R|pyg

s1/s2/hello.txt|jinja:
    code
"""



from dexy.common import OrderedDict
from dexy.data import Data
from dexy.doc import Doc
from dexy.exceptions import UserFeedback
from dexy.filter import DexyFilter
from dexy.tests.utils import wrap
from nose.tools import raises

@raises(UserFeedback)
def test_blank_alias():
    with wrap() as wrapper:
        doc = Doc("abc.txt|", wrapper=wrapper)
        doc.populate()

@raises(UserFeedback)
def test_blank_alias_exception():
    doc = Doc("foo")
    doc.filter_class_for_alias("")

def test_filter_class_for_alias():
    doc = Doc("foo")
    filter_class = doc.filter_class_for_alias("dexy")
    assert filter_class == DexyFilter

def test_output_is_data():
    with wrap() as wrapper:
        doc = Doc("abc.txt", contents="these are the contents", wrapper=wrapper)
        wrapper.run_docs(doc)
        assert isinstance(doc.output(), Data)

def test_create_virtual_initial_artifact():
    with wrap() as wrapper:
        doc = Doc("abc.txt", contents="these are the contents", wrapper=wrapper)
        wrapper.run_docs(doc)
        assert doc.children[0].__class__.__name__ == "InitialVirtualArtifact"
        assert doc.children[0].output_data.__class__.__name__ == "Generic"

def test_create_virtual_initial_artifact_with_dict():
    with wrap() as wrapper:
        od_contents = OrderedDict()
        od_contents['1'] = "these are the contents"
        doc = Doc("abc.txt", contents = od_contents, wrapper=wrapper)
        wrapper.run_docs(doc)
        assert doc.children[0].output_data.__class__.__name__ == "Sectioned"

def test_doc_setup():
    with wrap() as wrapper:
        with open("abc.txt", "w") as f:
            f.write("def")

        doc = Doc("abc.txt|dexy|dexy", wrapper=wrapper)

        doc.populate()

        for child in doc.children:
            child.setup()

        doc.setup()

        assert doc.key == "abc.txt|dexy|dexy"
        assert doc.name == "abc.txt"
        assert doc.filters == ["dexy", "dexy"]

        assert doc.children[0].key == "abc.txt"
        assert doc.children[1].key == "abc.txt|dexy"
        assert doc.children[2].key == "abc.txt|dexy|dexy"

        assert doc.children[0].__class__.__name__ == "InitialArtifact"
        assert doc.children[1].__class__.__name__ == "FilterArtifact"
        assert doc.children[2].__class__.__name__ == "FilterArtifact"

        assert not hasattr(doc.children[0], 'next_filter_alias')
        assert doc.children[1].next_filter_alias == "dexy"
        assert doc.children[2].next_filter_alias == None

        assert not doc.children[0].prior
        assert doc.children[1].prior.__class__.__name__ == "InitialArtifact"
        assert doc.children[2].prior.__class__.__name__ == "FilterArtifact"

        assert not doc.children[0].prior
        assert doc.children[1].prior.key == "abc.txt"
        assert doc.children[2].prior.key == "abc.txt|dexy"

